@@ -12,10 +12,10 @@ module Rake
 
   CommandLineOptionError = Class.new(StandardError)
 
-  ######################################################################
+  ##
   # Rake main application object.  When invoking +rake+ from the
   # command line, a Rake::Application object is created and run.
-  #
+
   class Application
     include TaskManager
     include TraceOutput
@@ -35,7 +35,12 @@ module Rake
     # List of the top level task names (task names from the command line).
     attr_reader :top_level_tasks
 
-    DEFAULT_RAKEFILES = ['rakefile', 'Rakefile', 'rakefile.rb', 'Rakefile.rb'].freeze
+    DEFAULT_RAKEFILES = [
+      'rakefile',
+      'Rakefile',
+      'rakefile.rb',
+      'Rakefile.rb'
+    ].freeze
 
     # Initialize a Rake::Application object.
     def initialize
@@ -79,7 +84,7 @@ module Rake
       standard_exception_handling do
         @name = app_name
         handle_options
-        collect_tasks
+        collect_command_line_tasks
       end
     end
 
@@ -112,10 +117,11 @@ module Rake
       thread_pool.join
       if options.job_stats
         stats = thread_pool.statistics
-        puts "Maximum active threads: #{stats[:max_active_threads]}"
-        puts "Total threads in play:  #{stats[:total_threads_in_play]}"
+        puts "Maximum active threads: #{stats[:max_active_threads]} + main"
+        puts "Total threads in play:  #{stats[:total_threads_in_play]} + main"
       end
-      ThreadHistoryDisplay.new(thread_pool.history).show if options.job_stats == :history
+      ThreadHistoryDisplay.new(thread_pool.history).show if
+        options.job_stats == :history
     end
 
     # Add a loader to handle imported files ending in the extension
@@ -132,62 +138,97 @@ module Rake
 
     # Return the thread pool used for multithreaded processing.
     def thread_pool             # :nodoc:
-      @thread_pool ||= ThreadPool.new(options.thread_pool_size||FIXNUM_MAX)
+      @thread_pool ||= ThreadPool.new(options.thread_pool_size || Rake.suggested_thread_count-1)
     end
 
-    # private ----------------------------------------------------------------
+    # internal ----------------------------------------------------------------
 
-    def invoke_task(task_string)
+    # Invokes a task with arguments that are extracted from +task_string+
+    def invoke_task(task_string) # :nodoc:
       name, args = parse_task_string(task_string)
       t = self[name]
       t.invoke(*args)
     end
 
-    def parse_task_string(string)
-      if string =~ /^([^\[]+)(\[(.*)\])$/
-        name = $1
-        args = $3.split(/\s*,\s*/)
-      else
-        name = string
-        args = []
-      end
-      [name, args]
+    def parse_task_string(string) # :nodoc:
+      /^([^\[]+)(?:\[(.*)\])$/ =~ string.to_s
+
+      name           = $1
+      remaining_args = $2
+
+      return string, [] unless name
+      return name,   [] if     remaining_args.empty?
+
+      args = []
+
+      begin
+        /((?:[^\\,]|\\.)*?)\s*(?:,\s*(.*))?$/ =~ remaining_args
+
+        remaining_args = $2
+        args << $1.gsub(/\\(.)/, '\1')
+      end while remaining_args
+
+      return name, args
     end
 
     # Provide standard exception handling for the given block.
-    def standard_exception_handling
-      begin
-        yield
-      rescue SystemExit => ex
-        # Exit silently with current status
-        raise
-      rescue OptionParser::InvalidOption => ex
-        $stderr.puts ex.message
-        exit(false)
-      rescue Exception => ex
-        # Exit with error message
-        display_error_message(ex)
-        exit_because_of_exception(ex)
-      end
+    def standard_exception_handling # :nodoc:
+      yield
+    rescue SystemExit
+      # Exit silently with current status
+      raise
+    rescue OptionParser::InvalidOption => ex
+      $stderr.puts ex.message
+      exit(false)
+    rescue Exception => ex
+      # Exit with error message
+      display_error_message(ex)
+      exit_because_of_exception(ex)
     end
 
     # Exit the program because of an unhandle exception.
     # (may be overridden by subclasses)
-    def exit_because_of_exception(ex)
+    def exit_because_of_exception(ex) # :nodoc:
       exit(false)
     end
 
     # Display the error message that caused the exception.
-    def display_error_message(ex)
+    def display_error_message(ex) # :nodoc:
       trace "#{name} aborted!"
-      trace ex.message
+      display_exception_details(ex)
+      trace "Tasks: #{ex.chain}" if has_chain?(ex)
+      trace "(See full trace by running task with --trace)" unless
+         options.backtrace
+    end
+
+    def display_exception_details(ex) # :nodoc:
+      seen = Thread.current[:rake_display_exception_details_seen] ||= []
+      return if seen.include? ex
+      seen << ex
+
+      display_exception_message_details(ex)
+      display_exception_backtrace(ex)
+      display_exception_details(ex.cause) if has_cause?(ex)
+    end
+
+    def has_cause?(ex) # :nodoc:
+      ex.respond_to?(:cause) && ex.cause
+    end
+
+    def display_exception_message_details(ex) # :nodoc:
+      if ex.instance_of?(RuntimeError)
+        trace ex.message
+      else
+        trace "#{ex.class.name}: #{ex.message}"
+      end
+    end
+
+    def display_exception_backtrace(ex) # :nodoc:
       if options.backtrace
         trace ex.backtrace.join("\n")
       else
         trace Backtrace.collapse(ex.backtrace).join("\n")
       end
-      trace "Tasks: #{ex.chain}" if has_chain?(ex)
-      trace "(See full trace by running task with --trace)" unless options.backtrace
     end
 
     # Warn about deprecated usage.
@@ -195,7 +236,7 @@ module Rake
     # Example:
     #    Rake.application.deprecate("import", "Rake.import", caller.first)
     #
-    def deprecate(old_usage, new_usage, call_site)
+    def deprecate(old_usage, new_usage, call_site) # :nodoc:
       unless options.ignore_deprecate
         $stderr.puts "WARNING: '#{old_usage}' is deprecated.  " +
           "Please use '#{new_usage}' instead.\n" +
@@ -204,14 +245,14 @@ module Rake
     end
 
     # Does the exception have a task invocation chain?
-    def has_chain?(exception)
+    def has_chain?(exception) # :nodoc:
       exception.respond_to?(:chain) && exception.chain
     end
     private :has_chain?
 
     # True if one of the files in RAKEFILES is in the current directory.
     # If a match is found, it is copied into @rakefile.
-    def have_rakefile
+    def have_rakefile # :nodoc:
       @rakefiles.each do |fn|
         if File.exist?(fn)
           others = FileList.glob(fn, File::FNM_CASEFOLD)
@@ -224,34 +265,40 @@ module Rake
     end
 
     # True if we are outputting to TTY, false otherwise
-    def tty_output?
+    def tty_output? # :nodoc:
       @tty_output
     end
 
     # Override the detected TTY output state (mostly for testing)
-    def tty_output=( tty_output_state )
+    def tty_output=(tty_output_state) # :nodoc:
       @tty_output = tty_output_state
     end
 
     # We will truncate output if we are outputting to a TTY or if we've been
     # given an explicit column width to honor
-    def truncate_output?
+    def truncate_output? # :nodoc:
       tty_output? || @terminal_columns.nonzero?
     end
 
     # Display the tasks and comments.
-    def display_tasks_and_comments
+    def display_tasks_and_comments # :nodoc:
       displayable_tasks = tasks.select { |t|
-        (options.show_all_tasks || t.comment) && t.name =~ options.show_task_pattern
+        (options.show_all_tasks || t.comment) &&
+          t.name =~ options.show_task_pattern
       }
       case options.show_tasks
       when :tasks
-        width = displayable_tasks.collect { |t| t.name_with_args.length }.max || 10
-        max_column = truncate_output? ? terminal_width - name.size - width - 7 : nil
+        width = displayable_tasks.map { |t| t.name_with_args.length }.max || 10
+        if truncate_output?
+          max_column = terminal_width - name.size - width - 7
+        else
+          max_column = nil
+        end
 
         displayable_tasks.each do |t|
-          printf "#{name} %-#{width}s  # %s\n",
-            t.name_with_args, max_column ? truncate(t.comment, max_column) : t.comment
+          printf("#{name} %-#{width}s  # %s\n",
+            t.name_with_args,
+            max_column ? truncate(t.comment, max_column) : t.comment)
         end
       when :describe
         displayable_tasks.each do |t|
@@ -265,7 +312,7 @@ module Rake
       when :lines
         displayable_tasks.each do |t|
           t.locations.each do |loc|
-            printf "#{name} %-30s %s\n",t.name_with_args, loc
+            printf "#{name} %-30s %s\n", t.name_with_args, loc
           end
         end
       else
@@ -273,7 +320,7 @@ module Rake
       end
     end
 
-    def terminal_width
+    def terminal_width # :nodoc:
       if @terminal_columns.nonzero?
         result = @terminal_columns
       else
@@ -285,50 +332,51 @@ module Rake
     end
 
     # Calculate the dynamic width of the
-    def dynamic_width
+    def dynamic_width # :nodoc:
       @dynamic_width ||= (dynamic_width_stty.nonzero? || dynamic_width_tput)
     end
 
-    def dynamic_width_stty
+    def dynamic_width_stty # :nodoc:
       %x{stty size 2>/dev/null}.split[1].to_i
     end
 
-    def dynamic_width_tput
+    def dynamic_width_tput # :nodoc:
       %x{tput cols 2>/dev/null}.to_i
     end
 
-    def unix?
-      RbConfig::CONFIG['host_os'] =~ /(aix|darwin|linux|(net|free|open)bsd|cygwin|solaris|irix|hpux)/i
+    def unix? # :nodoc:
+      RbConfig::CONFIG['host_os'] =~
+        /(aix|darwin|linux|(net|free|open)bsd|cygwin|solaris|irix|hpux)/i
     end
 
-    def windows?
+    def windows? # :nodoc:
       Win32.windows?
     end
 
-    def truncate(string, width)
+    def truncate(string, width) # :nodoc:
       if string.nil?
         ""
       elsif string.length <= width
         string
       else
-        ( string[0, width-3] || "" ) + "..."
+        (string[0, width - 3] || "") + "..."
       end
     end
 
     # Display the tasks and prerequisites
-    def display_prerequisites
+    def display_prerequisites # :nodoc:
       tasks.each do |t|
         puts "#{name} #{t.name}"
         t.prerequisites.each { |pre| puts "    #{pre}" }
       end
     end
 
-    def trace(*strings)
+    def trace(*strings) # :nodoc:
       options.trace_output ||= $stderr
       trace_on(options.trace_output, *strings)
     end
 
-    def sort_options(options)
+    def sort_options(options) # :nodoc:
       options.sort_by { |opt|
         opt.select { |o| o =~ /^-/ }.map { |o| o.downcase }.sort.reverse
       }
@@ -337,31 +385,42 @@ module Rake
 
     # A list of all the standard options used in rake, suitable for
     # passing to OptionParser.
-    def standard_rake_options
+    def standard_rake_options # :nodoc:
       sort_options(
         [
-          ['--all', '-A', "Show all tasks, even uncommented ones",
+          ['--all', '-A',
+            "Show all tasks, even uncommented ones (in combination with -T or -D)",
             lambda { |value|
               options.show_all_tasks = value
             }
           ],
-          ['--backtrace=[OUT]', "Enable full backtrace.  OUT can be stderr (default) or stdout.",
+          ['--backtrace=[OUT]',
+            "Enable full backtrace.  OUT can be stderr (default) or stdout.",
             lambda { |value|
               options.backtrace = true
               select_trace_output(options, 'backtrace', value)
             }
           ],
-          ['--comments', "Show commented tasks only",
+          ['--build-all', '-B',
+           "Build all prerequisites, including those which are up-to-date.",
+           lambda { |value|
+             options.build_all = true
+           }
+          ],
+          ['--comments',
+            "Show commented tasks only",
             lambda { |value|
               options.show_all_tasks = !value
             }
           ],
-          ['--describe', '-D [PATTERN]', "Describe the tasks (matching optional PATTERN), then exit.",
+          ['--describe', '-D [PATTERN]',
+            "Describe the tasks (matching optional PATTERN), then exit.",
             lambda { |value|
               select_tasks_to_show(options, :describe, value)
             }
           ],
-          ['--dry-run', '-n', "Do a dry run without executing actions.",
+          ['--dry-run', '-n',
+            "Do a dry run without executing actions.",
             lambda { |value|
               Rake.verbose(true)
               Rake.nowrite(true)
@@ -369,28 +428,43 @@ module Rake
               options.trace = true
             }
           ],
-          ['--execute',  '-e CODE', "Execute some Ruby code and exit.",
+          ['--execute', '-e CODE',
+            "Execute some Ruby code and exit.",
             lambda { |value|
               eval(value)
               exit
             }
           ],
-          ['--execute-print',  '-p CODE', "Execute some Ruby code, print the result, then exit.",
+          ['--execute-print', '-p CODE',
+            "Execute some Ruby code, print the result, then exit.",
             lambda { |value|
               puts eval(value)
               exit
             }
           ],
           ['--execute-continue',  '-E CODE',
-            "Execute some Ruby code, then continue with normal task processing.",
+            "Execute some Ruby code, " +
+            "then continue with normal task processing.",
             lambda { |value| eval(value) }
           ],
           ['--jobs',  '-j [NUMBER]',
-            "Specifies the maximum number of tasks to execute in parallel. (default:2)",
-            lambda { |value| options.thread_pool_size = [(value || 2).to_i,2].max }
+            "Specifies the maximum number of tasks to execute in parallel. " +
+            "(default is number of CPU cores + 4)",
+            lambda { |value|
+              if value.nil? || value == ''
+                value = FIXNUM_MAX
+              elsif value =~ /^\d+$/
+                value = value.to_i
+              else
+                value = Rake.suggested_thread_count
+              end
+              value = 1 if value < 1
+              options.thread_pool_size = value - 1
+            }
           ],
           ['--job-stats [LEVEL]',
-            "Display job statistics. LEVEL=history displays a complete job list",
+            "Display job statistics. " +
+            "LEVEL=history displays a complete job list",
             lambda { |value|
               if value =~ /^history/i
                 options.job_stats = :history
@@ -399,22 +473,28 @@ module Rake
               end
             }
           ],
-          ['--libdir', '-I LIBDIR', "Include LIBDIR in the search path for required modules.",
+          ['--libdir', '-I LIBDIR',
+            "Include LIBDIR in the search path for required modules.",
             lambda { |value| $:.push(value) }
           ],
-          ['--multitask', '-m', "Treat all tasks as multitasks.",
+          ['--multitask', '-m',
+            "Treat all tasks as multitasks.",
             lambda { |value| options.always_multitask = true }
           ],
-          ['--no-search', '--nosearch', '-N', "Do not search parent directories for the Rakefile.",
+          ['--no-search', '--nosearch',
+            '-N', "Do not search parent directories for the Rakefile.",
             lambda { |value| options.nosearch = true }
           ],
-          ['--prereqs', '-P', "Display the tasks and dependencies, then exit.",
+          ['--prereqs', '-P',
+            "Display the tasks and dependencies, then exit.",
             lambda { |value| options.show_prereqs = true }
           ],
-          ['--quiet', '-q', "Do not log messages to standard output.",
+          ['--quiet', '-q',
+            "Do not log messages to standard output.",
             lambda { |value| Rake.verbose(false) }
           ],
-          ['--rakefile', '-f [FILE]', "Use FILE as the rakefile.",
+          ['--rakefile', '-f [FILENAME]',
+            "Use FILENAME as the rakefile to search for.",
             lambda { |value|
               value ||= ''
               @rakefiles.clear
@@ -422,10 +502,14 @@ module Rake
             }
           ],
           ['--rakelibdir', '--rakelib', '-R RAKELIBDIR',
-            "Auto-import any .rake files in RAKELIBDIR. (default is 'rakelib')",
-            lambda { |value| options.rakelib = value.split(File::PATH_SEPARATOR) }
+            "Auto-import any .rake files in RAKELIBDIR. " +
+            "(default is 'rakelib')",
+            lambda { |value|
+              options.rakelib = value.split(File::PATH_SEPARATOR)
+            }
           ],
-          ['--require', '-r MODULE', "Require MODULE before executing rakefile.",
+          ['--require', '-r MODULE',
+            "Require MODULE before executing rakefile.",
             lambda { |value|
               begin
                 require value
@@ -438,34 +522,45 @@ module Rake
               end
             }
           ],
-          ['--rules', "Trace the rules resolution.",
+          ['--rules',
+            "Trace the rules resolution.",
             lambda { |value| options.trace_rules = true }
           ],
-          ['--silent', '-s', "Like --quiet, but also suppresses the 'in directory' announcement.",
+          ['--silent', '-s',
+            "Like --quiet, but also suppresses the " +
+            "'in directory' announcement.",
             lambda { |value|
               Rake.verbose(false)
               options.silent = true
             }
           ],
-          ['--suppress-backtrace PATTERN', "Suppress backtrace lines matching regexp PATTERN. Ignored if --trace is on.",
+          ['--suppress-backtrace PATTERN',
+            "Suppress backtrace lines matching regexp PATTERN. " +
+            "Ignored if --trace is on.",
             lambda { |value|
               options.suppress_backtrace_pattern = Regexp.new(value)
             }
           ],
           ['--system',  '-g',
-            "Using system wide (global) rakefiles (usually '~/.rake/*.rake').",
+            "Using system wide (global) rakefiles " +
+            "(usually '~/.rake/*.rake').",
             lambda { |value| options.load_system = true }
           ],
           ['--no-system', '--nosystem', '-G',
-            "Use standard project Rakefile search paths, ignore system wide rakefiles.",
+            "Use standard project Rakefile search paths, " +
+            "ignore system wide rakefiles.",
             lambda { |value| options.ignore_system = true }
           ],
-          ['--tasks', '-T [PATTERN]', "Display the tasks (matching optional PATTERN) with descriptions, then exit.",
+          ['--tasks', '-T [PATTERN]',
+            "Display the tasks (matching optional PATTERN) " +
+            "with descriptions, then exit.",
             lambda { |value|
               select_tasks_to_show(options, :tasks, value)
             }
           ],
-          ['--trace=[OUT]', '-t', "Turn on invoke/execute tracing, enable full backtrace. OUT can be stderr (default) or stdout.",
+          ['--trace=[OUT]', '-t',
+            "Turn on invoke/execute tracing, enable full backtrace. " +
+            "OUT can be stderr (default) or stdout.",
             lambda { |value|
               options.trace = true
               options.backtrace = true
@@ -473,22 +568,26 @@ module Rake
               Rake.verbose(true)
             }
           ],
-          ['--verbose', '-v', "Log message to standard output.",
+          ['--verbose', '-v',
+            "Log message to standard output.",
             lambda { |value| Rake.verbose(true) }
           ],
-          ['--version', '-V', "Display the program version.",
+          ['--version', '-V',
+            "Display the program version.",
             lambda { |value|
               puts "rake, version #{RAKEVERSION}"
               exit
             }
           ],
-          ['--where', '-W [PATTERN]', "Describe the tasks (matching optional PATTERN), then exit.",
+          ['--where', '-W [PATTERN]',
+            "Describe the tasks (matching optional PATTERN), then exit.",
             lambda { |value|
               select_tasks_to_show(options, :lines, value)
               options.show_all_tasks = true
             }
           ],
-          ['--no-deprecation-warnings', '-X', "Disable the deprecation warnings.",
+          ['--no-deprecation-warnings', '-X',
+            "Disable the deprecation warnings.",
             lambda { |value|
               options.ignore_deprecate = true
             }
@@ -496,14 +595,14 @@ module Rake
         ])
     end
 
-    def select_tasks_to_show(options, show_tasks, value)
+    def select_tasks_to_show(options, show_tasks, value) # :nodoc:
       options.show_tasks = show_tasks
       options.show_task_pattern = Regexp.new(value || '')
       Rake::TaskManager.record_task_metadata = true
     end
     private :select_tasks_to_show
 
-    def select_trace_output(options, trace_option, value)
+    def select_trace_output(options, trace_option, value) # :nodoc:
       value = value.strip unless value.nil?
       case value
       when 'stdout'
@@ -511,18 +610,19 @@ module Rake
       when 'stderr', nil
         options.trace_output = $stderr
       else
-        fail CommandLineOptionError, "Unrecognized --#{trace_option} option '#{value}'"
+        fail CommandLineOptionError,
+          "Unrecognized --#{trace_option} option '#{value}'"
       end
     end
     private :select_trace_output
 
     # Read and handle the command line options.
-    def handle_options
+    def handle_options # :nodoc:
       options.rakelib = ['rakelib']
       options.trace_output = $stderr
 
       OptionParser.new do |opts|
-        opts.banner = "rake [-f rakefile] {options} targets..."
+        opts.banner = "#{Rake.application.name} [-f rakefile] {options} targets..."
         opts.separator ""
         opts.separator "Options are ..."
 
@@ -538,7 +638,7 @@ module Rake
 
     # Similar to the regular Ruby +require+ command, but will check
     # for *.rake files in addition to *.rb files.
-    def rake_require(file_name, paths=$LOAD_PATH, loaded=$")
+    def rake_require(file_name, paths=$LOAD_PATH, loaded=$") # :nodoc:
       fn = file_name + ".rake"
       return false if loaded.include?(fn)
       paths.each do |path|
@@ -552,13 +652,11 @@ module Rake
       fail LoadError, "Can't find #{file_name}"
     end
 
-    def find_rakefile_location
+    def find_rakefile_location # :nodoc:
       here = Dir.pwd
-      while ! (fn = have_rakefile)
+      until (fn = have_rakefile)
         Dir.chdir("..")
-        if Dir.pwd == here || options.nosearch
-          return nil
-        end
+        return nil if Dir.pwd == here || options.nosearch
         here = Dir.pwd
       end
       [fn, here]
@@ -566,7 +664,7 @@ module Rake
       Dir.chdir(Rake.original_dir)
     end
 
-    def print_rakefile_directory(location)
+    def print_rakefile_directory(location) # :nodoc:
       $stderr.puts "(in #{Dir.pwd})" unless
         options.silent or original_dir == location
     end
@@ -586,7 +684,8 @@ module Rake
         @rakefile = rakefile
         Dir.chdir(location)
         print_rakefile_directory(location)
-        Rake.load_rakefile(File.expand_path(@rakefile)) if @rakefile && @rakefile != ''
+        Rake.load_rakefile(File.expand_path(@rakefile)) if
+          @rakefile && @rakefile != ''
         options.rakelib.each do |rlib|
           glob("#{rlib}/*.rake") do |name|
             add_import name
@@ -596,13 +695,13 @@ module Rake
       load_imports
     end
 
-    def glob(path, &block)
+    def glob(path, &block) # :nodoc:
       FileList.glob(path.gsub("\\", '/')).each(&block)
     end
     private :glob
 
     # The directory path containing the system wide rakefiles.
-    def system_dir
+    def system_dir # :nodoc:
       @system_dir ||=
         begin
           if ENV['RAKE_SYSTEM']
@@ -628,7 +727,7 @@ module Rake
     # Collect the list of tasks on the command line.  If no tasks are
     # given, return a list containing only the default task.
     # Environmental assignments are processed at this time as well.
-    def collect_tasks
+    def collect_command_line_tasks # :nodoc:
       @top_level_tasks = []
       ARGV.each do |arg|
         if arg =~ /^(\w+)=(.*)$/m
@@ -642,31 +741,34 @@ module Rake
 
     # Default task name ("default").
     # (May be overridden by subclasses)
-    def default_task_name
+    def default_task_name # :nodoc:
       "default"
     end
 
     # Add a file to the list of files to be imported.
-    def add_import(fn)
+    def add_import(fn) # :nodoc:
       @pending_imports << fn
     end
 
     # Load the pending list of imported files.
-    def load_imports
+    def load_imports # :nodoc:
       while fn = @pending_imports.shift
         next if @imported.member?(fn)
-        if fn_task = lookup(fn)
-          fn_task.invoke
-        end
+        fn_task = lookup(fn) and fn_task.invoke
         ext = File.extname(fn)
         loader = @loaders[ext] || @default_loader
         loader.load(fn)
+        if fn_task = lookup(fn) and fn_task.needed?
+          fn_task.reenable
+          fn_task.invoke
+          loader.load(fn)
+        end
         @imported << fn
       end
     end
 
-    def rakefile_location(backtrace=caller)
-      backtrace.map { |t| t[/([^:]+):/,1] }
+    def rakefile_location(backtrace=caller) # :nodoc:
+      backtrace.map { |t| t[/([^:]+):/, 1] }
 
       re = /^#{@rakefile}$/
       re = /#{re.source}/i if windows?
